from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from django.contrib.auth import authenticate
from django import forms

from evotecore import widgets

from .models import User


class EvoteAuthenticationForm(forms.Form):
    """
    Base class for authenticating users. Extend this to get a form that accepts
    username/password logins.
    """
    email = forms.EmailField(
        label=_("Email"), max_length=254,
        widget=forms.EmailInput(
            attrs={'autofocus': True,
                   'placeholder': _("Email"),
                   'autocomplete': 'off'}))
    password = forms.CharField(
        label=_("Password"), strip=False,
        widget=forms.PasswordInput(
            attrs={'placeholder': _("Password"),
                   'autocomplete': 'off'}))

    error_messages = {
        'invalid_login': _(
            "Please enter a correct email and password. Note that both "
            "fields may be case-sensitive."
        ),
        'inactive': _("This account is inactive."),
    }

    def __init__(self, *args, **kwargs):
        """ """
        self.user_cache = None
        super(EvoteAuthenticationForm, self).__init__(*args, **kwargs)

    def clean(self):
        email = self.cleaned_data.get('email')
        password = self.cleaned_data.get('password')
        if email is not None and password:
            self.user_cache = authenticate(email=email, password=password)
            if self.user_cache is None:
                raise forms.ValidationError(
                    self.error_messages['invalid_login'],
                    code='invalid_login'
                )
            else:
                self.confirm_login_allowed(
                    self.user_cache, cleaned_data=self.cleaned_data)
        return self.cleaned_data

    def confirm_login_allowed(self, user, cleaned_data=None):
        """
        Controls whether the given User may log in. This is a policy setting,
        independent of end-user authentication. This default behavior is to
        allow login by active users, and reject login by inactive users.
        If the given user cannot log in, this method should raise a
        ``forms.ValidationError``.
        If the given user may log in, this method should return None.
        """
        if not user.is_active:
            raise forms.ValidationError(
                self.error_messages['inactive'],
                code='inactive',
            )

    def get_user_id(self):
        if self.user_cache:
            return self.user_cache.id
        return None

    def get_user(self):
        return self.user_cache


class PersonalDataForm(forms.ModelForm):
    """ """
    error_messages = {
        'empty_notice_email': _("Please provide the notice email"),
        'empty_notice_phone': _("Please provide the notice phone"),
        'invalid_notice_email': _("Enter a valid email address"),
        'invalid_notice_phone': _("Enter a valid phone number"),
    }

    class Meta:
        model = User
        fields = ('notice_method', 'notice_phone', 'notice_email')
        widgets = {
            'notice_method': widgets.Select
        }

    def __init__(self, *args, **kwargs):
        self.base_url = kwargs.pop("base_url")
        super(PersonalDataForm, self).__init__(*args, **kwargs)
        self.fields['notice_phone'].widget.attrs.update({
            'placeholder': "+351900000000"})
        self.fields['notice_phone'].error_messages.update({
            'invalid': self.error_messages['invalid_notice_phone']
        })
        self.fields['notice_email'].error_messages.update({
            'invalid': self.error_messages['invalid_notice_email']
        })

    def clean_notice_email(self):
        notice_email = self.cleaned_data.get('notice_email', None)
        if notice_email and notice_email != self.instance.notice_email:
            self.instance.set_pending_email(notice_email, self.base_url)
        elif notice_email == '':
            return notice_email
        return self.instance.notice_email

    def clean(self):
        super(PersonalDataForm, self).clean()
        notice_method = self.cleaned_data.get('notice_method', None)
        notice_email = self.cleaned_data.get('notice_email', None)
        notice_phone = self.cleaned_data.get('notice_phone', None)
        if notice_method == User.NoticeMethod.EMAIL and \
                not self.instance.email and not notice_email:
            self.add_error('notice_email',
                           self.error_messages['empty_notice_email'])
        if notice_method == User.NoticeMethod.SMS and \
                not self.instance.primary_phone and not notice_phone:
            self.add_error('notice_phone',
                           self.error_messages['empty_notice_phone'])
        return self.cleaned_data
