from __future__ import unicode_literals

from django.test import TestCase
from django import forms

from evoteusers.tests.utils import UserTestUtils
from evoteusers.forms import EvoteAuthenticationForm, PersonalDataForm


class EvoteAuthenticationFormTest(TestCase, UserTestUtils):
    """ """

    def setUp(self):
        self.credentials = {'email': 'dev@eurotux.com', 'password': 'password'}
        self.user = self.create_user(**self.credentials)

    def test_default(self):
        form = EvoteAuthenticationForm(data=self.credentials)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.get_user(), self.user)

    def test_empty(self):
        form = EvoteAuthenticationForm(data={})
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            "email": ["This field is required."],
            "password": ["This field is required."]
        })

    def test_with_invalid_credentials(self):
        self.credentials['password'] = 'wrong'
        form = EvoteAuthenticationForm(data=self.credentials)
        self.assertFalse(form.is_valid())
        self.assertIsNone(form.get_user())

    def test_with_inactive_user(self):
        self.user.is_active = False
        self.user.save()
        form = EvoteAuthenticationForm(data=self.credentials)
        self.assertFalse(form.is_valid())
        with self.assertRaisesMessage(
                forms.ValidationError, 'This account is inactive.'):
            form.confirm_login_allowed(self.user)

    def test_get_user_id_with_valid_authentication(self):
        form = EvoteAuthenticationForm(data=self.credentials)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.get_user_id(), self.user.id)

    def test_get_user_id_with_invalid_authentication(self):
        self.credentials['password'] = 'wrong'
        form = EvoteAuthenticationForm(data=self.credentials)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.get_user_id(), None)


class PersonalDataFormTest(TestCase, UserTestUtils):
    """ """

    def setUp(self):
        self.login()

    def test_clean_notice_email_same(self):
        """ Test new notice email change (same) """
        self.user.notice_email = 'dev@eurotux.com'
        self.user.save(update_fields=['notice_email'])
        data = {
            'photo': self.user.photo,
            'name': self.user.email,
            'email': self.user.email,
            'primary_phone': self.user.primary_phone,
            'notice_method': self.user.notice_method,
            'notice_phone': self.user.notice_phone,
            'notice_email': self.user.notice_email,
        }
        form = PersonalDataForm(data=data, instance=self.user,
                                base_url='http://base.com/')
        form.is_valid()
        self.assertEqual(form.clean_notice_email(), self.user.notice_email)

    def test_clean_notice_email_empty(self):
        """ Test new notice email change (empty) """
        self.user.notice_email = 'dev@eurotux.com'
        self.user.save(update_fields=['notice_email'])
        data = {
            'photo': self.user.photo,
            'name': self.user.email,
            'email': self.user.email,
            'primary_phone': self.user.primary_phone,
            'notice_method': self.user.notice_method,
            'notice_phone': self.user.notice_phone,
            'notice_email': '',
        }
        form = PersonalDataForm(data=data, instance=self.user,
                                base_url='http://base.com/')
        form.is_valid()
        self.assertEqual(form.clean_notice_email(), '')

    def test_clean_notice_email_to_pending(self):
        """ Test new notice email change (to pending) """
        self.user.notice_email = 'dev@eurotux.com'
        self.user.save(update_fields=['notice_email'])
        data = {
            'photo': self.user.photo,
            'name': self.user.email,
            'email': self.user.email,
            'primary_phone': self.user.primary_phone,
            'notice_method': self.user.notice_method,
            'notice_phone': self.user.notice_phone,
            'notice_email': 'unit@eurotux.com',
        }
        form = PersonalDataForm(data=data, instance=self.user,
                                base_url='http://base.com/')
        form.is_valid()
        self.assertEqual(form.clean_notice_email(), self.user.notice_email)
        self.assertEqual(self.user.pending_email, 'unit@eurotux.com')
