from __future__ import unicode_literals

import requests
import base64
import json
import logging

from django.conf import settings
from django.apps import apps

from rest_framework.renderers import JSONRenderer
from evotevotings.admin.utils import get_certificates


logger = logging.getLogger('evote.services')


class BaseRouter(object):

    endpoint = None

    def __init__(self):
        # Get all endpoints from the settings
        services_endpoints = settings.EVOTE_SERVICES_ENDPOINTS
        # set the anonymizer service base url
        self.endpoint = services_endpoints.get(self.endpoint_name)


class ASrouter(BaseRouter):
    """Helper for Anonymizer Service communication handling"""
    endpoint_name = 'anonymizer'

    def build_url(self, pk, path):
        """Method to build base endpoint URL for a given election ID
        """
        return "{}election/{}/{}/".format(self.endpoint, pk, path)

    def anonymize_vote(self, election_id, data):
        """Method to insert a vote into the AS database"""
        # Build the endpoint URL
        url = self.build_url(election_id, "anonymizevote")

        # Connect to the endpoint
        response = requests.post(
            url,
            data=json.dumps(data),
            headers={'content-type': 'application/json'},
            verify=False,
            timeout=None)

        if response.status_code == 200:
            # Parse the response
            json_response = response.json()
            # Expected boolean response
            return json_response
        else:
            logger.critical(
                "Anonymize vote: url=%s status_code=%s",
                url, response.status_code)
        return False

    def initialize_election(self, election_id, voters_nr):
        """Method to handle election initialization"""
        # Build the endpoint URL
        url = self.build_url(election_id, "initialize")
        # Connect to the endpoint
        response = requests.post(
            url,
            data=json.dumps({'voters_nr': voters_nr}),
            timeout=None,
            headers={'content-type': 'application/json'},
            verify=False)
        # Parse the response
        json_response = response.json()
        # Error found
        if response.status_code != 200 or json_response['error'] is True:
            logger.critical(
                "Initialize election: url=%s status_code=%s response=%r",
                url, response.status_code, json_response)
            return False
        return True

    def close_election(self, election_id):
        """Method to handle election closing"""
        # Build the endpoint URL
        url = self.build_url(election_id, "close")

        # Connect to the endpoint
        response = requests.post(
            url, data=json.dumps({}),
            headers={'content-type': 'application/json'},
            verify=False)

        # Parse the response
        json_response = response.json()

        # Error found, raise exception
        if response.status_code != 200:
            logger.error("Close election: url=%s status_code=%s response=%r",
                         url, response.status_code, json_response[0])
            return False
        return True

    def list_references(self, election_id, page=1, search=None):
        """ Method to handle references search and listing """
        # List to hold the returned references
        references = []
        more = False

        # Build the endpoint URL
        url = self.build_url(election_id, "references")
        # Create the required GET data
        request_data = {'page': page}
        if search is not None:
            request_data['unique_vote_bulletin'] = search

        # Connect to the endpoint
        response = requests.get(url, request_data, verify=False)
        reference_list = response.json()

        # Check for valid response
        if response.status_code == 200 and 'count' in reference_list:
            for reference in reference_list['results']:
                references.append(reference['unique_vote_bulletin'])

            # Check for next page
            if reference_list['next'] is not None:
                more = True
        else:
            logger.error("References: url=%s status_code=%d",
                         url, response.status_code)

        return (references, more)

    def partial_count(self, election_id):
        """ Method to get the current vote count """
        # Build the endpoint URL
        url = self.build_url(election_id, "count")

        count = 0
        # Connect to the endpoint
        response = requests.get(
            url,
            headers={'content-type': 'application/json'},
            verify=False)

        # Parse the response
        json_response = response.json()

        # Check for valid response
        if response.status_code == 200 and 'count' in json_response:
            count = json_response['count']
        else:
            logger.error("Partial count: url=%s status_code %d",
                         url, response.status_code)
        return count


class CSrouter(BaseRouter):
    """ Helper for Counter Service communication handling """
    # Counter service endpoints
    endpoint_name = 'counter'

    def build_url(self, pk, path):
        """Method to build base endpoint URL for a given election ID
        """
        return "{}scrutiny/{}/{}/".format(self.endpoint, pk, path)

    def get_election_model(self):
        return apps.get_model("evoteelections.Election")  # noqa

    def get_electoral_member_model(self):
        return apps.get_model("evoteprocesses.ElectoralMember")  # noqa

    def generate_results(self, election_id, process_id):
        """ Method to generate the election results """
        # Build the endpoint URL
        url = self.build_url(election_id, "generate")

        results = None

        # Fetch all data to be sent to the counter service
        secret_entries = get_certificates(election_id)
        # Election private key (from Election)
        elec_pvt_key = secret_entries['election_pvt_key']
        # Counter Service private key
        cs_pvt_key = secret_entries['cs_pvt_key']
        # Election management certificate
        sge_cert = secret_entries['sge']
        # Validator Service certificate
        vs_cert = secret_entries['vs']
        # Root (CA) certificate
        root_cert = secret_entries['root']

        # List of members shared secrets
        ElectoralMember = self.get_electoral_member_model()  # NOQA
        secret_list = ElectoralMember.objects\
            .filter(electoral_process_id=process_id)\
            .values_list('private_key', flat=True)

        # Build the request POST data
        # Encoding all certificates and private keys
        request_data = {
            'election_pvt_key': base64.b64encode(elec_pvt_key),
            'cs_pvt_key': base64.b64encode(cs_pvt_key),
            'vs_cert': base64.b64encode(vs_cert),
            'root_cert': base64.b64encode(root_cert),
            'sge_cert': base64.b64encode(sge_cert),
            'secret_list': secret_list,
            'elec_proc_id': process_id
        }

        response = requests.post(
            url,
            JSONRenderer().render(request_data),
            headers={'content-type': 'application/json'},
            verify=False
        )

        json_response = response.json()

        if response.status_code == 200:
            # Check for errors
            if json_response['error']:
                logger.critical("Scrutiny results: error message=%s",
                                json_response['message'])
            elif json_response['message'] is not False:
                results = json_response['message']
        else:
            logger.critical("Scrutiny results: url=%s status_code=%d",
                            url, response.status_code)
            results = json_response
        return results

    def check_results_status(self, election_id):
        """ Method to check the current scrutiny status """

        # Build the endpoint URL
        url = self.build_url(election_id, "status")

        # Return value
        status = None
        # Connect to the endpoint
        response = requests.get(
            url, headers={'content-type': 'application/json'},
            verify=False)

        # Parse the response
        json_response = response.json()

        # Check for valid response
        if response.status_code == 200 and \
                'result_status_choices' in json_response:
            status = int(json_response['result_status_choices'])
        else:
            logger.warning(
                "Check results status: url=%s status_code=%d",
                url, response.status_code)
        return status

    def clean_results(self, election_id):
        """ Method to clean all results associated to a given election ID.
        Returns boolean."""
        url = self.build_url(election_id, "clean")

        # Connect to the endpoint
        response = requests.get(url, verify=False)

        # Set the response code to return
        if response.status_code != 200:
            logger.error("Clean results: url=%s status_code=%d",
                         url, response.status_code)
            return False
        return True
