from __future__ import unicode_literals

import logging

from django.utils.translation import ugettext_lazy as _
from django.core.validators import validate_email
from django.core.exceptions import ValidationError
from django.apps import apps
from django.db import transaction

from evotelogs.registers import audit_register
from evotecore.importers import BaseImporter


logger = logging.getLogger('evote.importers')


class ElectoralRollImporter(BaseImporter):
    """ """
    field_names = ('personnel_number', 'name', 'email', 'category')
    BaseImporter.error_messages.update({
        'unique_personnel_number': _("Personnel number must be unique"),
        'unique_email': _("Email must be unique"),
        'invalid_email': _("Email invalid format"),
        'invalid_category': _("Voting weight not defined"),
    })
    category_slugs = []

    def __init__(self, election=None, delimiter=None):
        """ """
        super(ElectoralRollImporter, self).__init__(delimiter)
        if election:
            self.category_slugs = election.categories\
                .values_list('slug', flat=True)

    @transaction.atomic
    def run(self, electoral_roll, document):
        """ """
        data = self.get_csv_data(document, True)
        ElectionVoter = apps.get_model("evotevoters.ElectionVoter")  # noqa
        User = apps.get_model("evoteusers.User")  # noqa
        # unregister ElectionVoter to dont log the imported info
        audit_register.remove(ElectionVoter)
        audit_register.remove(User)
        logger.info("Electoral roll %d: importing election voter",
                    electoral_roll.pk)
        # do the import
        result = ElectionVoter.objects.bulk_import(electoral_roll, data)
        # register ElectionVoter again
        audit_register.add(ElectionVoter)
        audit_register.add(User)
        return result

    def get_validator(self):
        """ """
        validator = super(ElectoralRollImporter, self).get_validator()
        validator.add_unique_check(
            'personnel_number',
            message=self.error_messages['unique_personnel_number'])
        validator.add_unique_check(
            'email', message=self.error_messages['unique_email'])
        validator.add_value_predicate(
            'email', self._check_email,
            message=self.error_messages['invalid_email'])
        validator.add_value_predicate(
            'category', self._check_category,
            message=self.error_messages['invalid_category'])
        return validator

    def _check_category(self, value):
        """ """
        if value:
            return value in self.category_slugs
        return True

    def _check_email(self, value):
        try:
            validate_email(value)
        except ValidationError:
            return False
        return True
