# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from StringIO import StringIO
from csvvalidator import VALUE_PREDICATE_FALSE, UNIQUE_CHECK_FAILED

from django.test import TestCase

from evotevotings.voters.tests.utils import VotersTestUtils
from evotevotings.voters.importers import ElectoralRollImporter


class ElectoralRollImporterTestCase(VotersTestUtils, TestCase):
    """ """
    def setUp(self):
        """ """
        self.electoral_roll = self.create_electoral_roll()
        self.importer = ElectoralRollImporter(self.electoral_roll.election)
        self.header = "personnel_number;name;email;category\n"

    def _get_csv(self, lines=None):
        """ """
        csv = self.header
        if lines:
            csv += "\n".join(lines)
        csv = csv.encode('utf-8')
        return StringIO(csv)

    def test_validate_with_valid_header_fields(self):
        """ """
        csv = self._get_csv()
        result = self.importer.validate(csv)
        self.assertEqual(result, [])

    def test_validate_with_invalid_header_fields(self):
        """ """
        invalid_csv_header = StringIO("personnel_number;names;category")
        result = self.importer.validate(invalid_csv_header)
        self.assertEqual(result[0]['unexpected'], set(["names"]))
        self.assertEqual(result[0]['missing'], set(["name", "email"]))

    def test_validate_unique_email(self):
        """ """
        csv = self._get_csv([
            "123;test name;dev@eurotux.com;one",
            "124;test name;dev@eurotux.com;one"
        ])
        result = self.importer.validate(csv)
        self.assertEqual(result[0]['code'], UNIQUE_CHECK_FAILED)
        self.assertEqual(result[0]['key'], "email")

    def test_validate_unique_personnel_number(self):
        """ """
        csv = self._get_csv([
            "123;test name;dev1@eurotux.com;one",
            "123;test name;dev2@eurotux.com;one"
        ])
        result = self.importer.validate(csv)
        self.assertEqual(result[0]['code'], UNIQUE_CHECK_FAILED)
        self.assertEqual(result[0]['key'], "personnel_number")

    def test_validate_category(self):
        """ """
        csv = self._get_csv([
            "123;test name;dev@eurotux.com;two",
        ])
        result = self.importer.validate(csv)
        self.assertEqual(result[0]['code'], VALUE_PREDICATE_FALSE)
        self.assertEqual(result[0]['field'], "category")

    def test_validate_category_not_required(self):
        """ """
        csv = self._get_csv([
            "123;test name;dev@eurotux.com;",
        ])
        result = self.importer.validate(csv)
        self.assertEqual(result, [])

    def test_validate_email_format(self):
        """ """
        csv = self._get_csv([
            "123;test name;dev@eurotux.com;one",
            "123;test name;dev@eurotux;one",
        ])
        result = self.importer.validate(csv)
        self.assertEqual(result[0]['code'], VALUE_PREDICATE_FALSE)
        self.assertEqual(result[0]['field'], "email")

    def test_run(self):
        csv = self._get_csv([
            "1;test name;dev1@eurotux.com;one",
            "2;test name;dev2@eurotux.com;one",
        ])
        voters = self.importer.run(self.electoral_roll, csv)
        self.assertIsNotNone(voters)
        self.assertEqual(len(voters), 2)

    def test_run_with_special_characters(self):
        csv = self._get_csv([
            "1;test name é;dev1@eurotux.com;one",
            "2;test name ó;dev2@eurotux.com;one",
        ])
        voters = self.importer.run(self.electoral_roll, csv)
        self.assertIsNotNone(voters)
        self.assertEqual(len(voters), 2)

    def test_run_with_already_existent_data(self):
        """Test update and new insertion"""
        content = [
            "1;test name;dev1@eurotux.com;one",
            "2;test name;dev2@eurotux.com;one",
        ]
        # insert new voters
        self.importer.run(self.electoral_roll, self._get_csv(content))
        content.append("3;test name 3;dev3@eurotux.com;one")
        # insert new voter and update 2
        voters = self.importer.run(self.electoral_roll, self._get_csv(content))
        self.assertEqual(voters.count(),
                         self.electoral_roll.election_voters.count())
